/*
 * Copyright (c) 2002-2006, Network Appliance, Inc. All rights reserved.
 *
 * This Software is licensed under all of the following licenses:
 * 
 * 1) under the terms of the "Common Public License 1.0" a copy of which is
 *    in the file LICENSE.txt in the root directory. The license is also
 *    available from the Open Source Initiative, see 
 *    http://www.opensource.org/licenses/cpl.php.
 * 
 * 2) under the terms of the "The BSD License" a copy of which is in the file
 *    LICENSE2.txt in the root directory. The license is also available from
 *    the Open Source Initiative, see
 *    http://www.opensource.org/licenses/bsd-license.php.
 * 
 * 3) under the terms of the "GNU General Public License (GPL) Version 2" a
 *    copy of which is in the file LICENSE3.txt in the root directory. The
 *    license is also available from the Open Source Initiative, see
 *    http://www.opensource.org/licenses/gpl-license.php.
 * 
 * Licensee has the right to choose one of the above licenses.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain both the above copyright
 * notice and one of the license notices.
 * 
 * Redistributions in binary form must reproduce both the above copyright
 * notice, one of the license notices in the documentation
 * and/or other materials provided with the distribution.
 */

/***************************************************************
 *
 * HEADER: dat_vendor_specific.h
 *
 * PURPOSE:
 *
 * Description: Header file for "DAPL: Direct Access Programming
 *		Library, Version: 2.0"
 *
 * Mapping rules:
 *
 ***************************************************************/

#ifndef _DAT_VENDOR_SPECIFIC_H_
#define _DAT_VENDOR_SPECIFIC_H_



/* Vendor-specific extensions */

#if defined(_AMMASSO)

#elif defined(_BROADCOM)

#elif defined(_CISCO)

#elif defined(_IBM)

#elif defined(_INTEL)

#elif defined(_JNI)

#elif defined(_MELLANOX)

#elif defined(_MYRINET)

#elif defined(_NETEFFECT)

#elif defined(_QLOGIC)

#elif defined(_SILVERSTORM)

#elif defined(_VOLTAIRE)

#endif

#endif /* _DAT_VENDOR_SPECIFIC_H_ */
