/*
 * Copyright (c) 2007 Intel Corporation.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id$
 */


#include <oib_ver.h>

#define VER_FILETYPE			VFT_APP
#define VER_FILESUBTYPE			VFT2_UNKNOWN

#if DBG
#define VER_FILEDESCRIPTION_STR		"Simple DAPL/DAT svr/cli test Application (Debug)"
#define VER_INTERNALNAME_STR		"dtest2d.exe"
#define VER_ORIGINALFILENAME_STR	"dtest2d.exe"
#else
#define VER_FILEDESCRIPTION_STR		"Simple DAPL/DAT svr/cli test Application"
#define VER_INTERNALNAME_STR		"dtest2.exe"
#define VER_ORIGINALFILENAME_STR	"dtest2.exe"
#endif

#include <common.ver>
