#include "test/dapl_bpool.c"
#include "test/dapl_client.c"
#include "test/dapl_client_info.c"
#include "test/dapl_cnxn.c"
#include "test/dapl_execute.c"
#include "test/dapl_fft_connmgt.c"
#include "test/dapl_fft_endpoint.c"
#include "test/dapl_fft_hwconn.c"
#include "test/dapl_fft_mem.c"
#include "test/dapl_fft_pz.c"
#include "test/dapl_fft_queryinfo.c"
#include "test/dapl_fft_test.c"
#include "test/dapl_fft_util.c"
#include "test/dapl_limit.c"
#include "test/dapl_memlist.c"
#include "test/dapl_performance_client.c"
#include "test/dapl_performance_server.c"
#include "test/dapl_performance_stats.c"
#include "test/dapl_performance_util.c"
#include "test/dapl_quit_util.c"
#include "test/dapl_server.c"
#include "test/dapl_server_info.c"
#include "test/dapl_test_data.c"
#include "test/dapl_test_util.c"
#include "test/dapl_thread.c"
#include "test/dapl_transaction_stats.c"
#include "test/dapl_transaction_test.c"
#include "test/dapl_transaction_util.c"
#include "test/dapl_util.c"
#if defined(DAT_EXPORT_SYMBOLS)
#error DAT_EXPORT?
#endif
